@@ -1,3 +1,5 @@
+// Copyright 2018-2021 the Deno authors. All rights reserved. MIT license.
+
 use std::path::PathBuf;
 
 use deno_node_transform::Dependency;
@@ -8,6 +10,9 @@ mod integration;
 
 use integration::TestBuilder;
 
+use crate::integration::assert_identity_transforms;
+use crate::integration::assert_transforms;
+
 #[tokio::test]
 async fn transform_standalone_file() {
   let result = TestBuilder::new()
@@ -23,124 +28,143 @@ async fn transform_standalone_file() {
 
 #[tokio::test]
 async fn transform_deno_shim() {
-  let result = TestBuilder::new()
-    .with_loader(|loader| {
-      loader.add_local_file("/mod.ts", r#"Deno.readTextFile();"#);
-    })
-    .transform()
-    .await
-    .unwrap();
-
-  assert_files!(
-    result.main.files,
-    &[(
-      "mod.ts",
+  assert_transforms(vec![
+    (
+      "Deno.readTextFile();",
       concat!(
-        r#"import * as denoShim from "deno.ns";"#,
+        r#"import * as denoShim from "test-shim";"#,
         "\ndenoShim.Deno.readTextFile();"
-      )
-    )]
-  );
+      ),
+    ),
+    (
+      "const [test=Deno] = other;",
+      concat!(
+        r#"import * as denoShim from "test-shim";"#,
+        "\nconst [test=denoShim.Deno] = other;"
+      ),
+    ),
+    (
+      "const obj = { test: Deno };",
+      concat!(
+        r#"import * as denoShim from "test-shim";"#,
+        "\nconst obj = { test: denoShim.Deno };"
+      ),
+    ),
+    (
+      concat!(
+        "const decl01 = Blob;\n",
+        "const decl02 = crypto;\n",
+        "const decl03 = fetch;\n",
+        "const decl04 = File;\n",
+        "const decl05 = FormData;\n",
+        "const decl06 = Headers;\n",
+        "const decl07 = Request;\n",
+        "const decl08 = Response;\n",
+        "const decl09 = alert;\n",
+        "const decl10 = confirm;\n",
+        "const decl11: typeof prompt = prompt;\n",
+      ),
+      concat!(
+        r#"import * as denoShim from "test-shim";"#,
+        "\nconst decl01 = denoShim.Blob;\n",
+        "const decl02 = denoShim.crypto;\n",
+        "const decl03 = denoShim.fetch;\n",
+        "const decl04 = denoShim.File;\n",
+        "const decl05 = denoShim.FormData;\n",
+        "const decl06 = denoShim.Headers;\n",
+        "const decl07 = denoShim.Request;\n",
+        "const decl08 = denoShim.Response;\n",
+        "const decl09 = denoShim.alert;\n",
+        "const decl10 = denoShim.confirm;\n",
+        "const decl11: typeof denoShim.prompt = denoShim.prompt;\n",
+      ),
+    ),
+  ])
+  .await;
 }
 
 #[tokio::test]
 async fn no_transform_deno_ignored() {
-  let result = TestBuilder::new()
-    .with_loader(|loader| {
-      loader
-        .add_local_file("/mod.ts", "// deno-shim-ignore\nDeno.readTextFile();");
-    })
-    .transform()
-    .await
-    .unwrap();
-
-  assert_files!(
-    result.main.files,
-    &[("mod.ts", "// deno-shim-ignore\nDeno.readTextFile();",)]
-  );
+  assert_identity_transforms(vec!["// deno-shim-ignore\nDeno.readTextFile();"])
+    .await;
 }
 
 #[tokio::test]
 async fn transform_deno_shim_with_name_collision() {
-  let result = TestBuilder::new()
-    .with_loader(|loader| {
-      loader.add_local_file(
-        "/mod.ts",
-        r#"Deno.readTextFile(); const denoShim = {};"#,
-      );
-    })
-    .shim_package_name("test-shim")
-    .transform()
-    .await
-    .unwrap();
-
-  assert_files!(
-    result.main.files,
-    &[(
-      "mod.ts",
-      concat!(
-        r#"import * as denoShim1 from "test-shim";"#,
-        "\ndenoShim1.Deno.readTextFile(); const denoShim = {};"
-      )
-    )]
-  );
+  assert_transforms(vec![(
+    "Deno.readTextFile(); const denoShim = {};",
+    concat!(
+      r#"import * as denoShim1 from "test-shim";"#,
+      "\ndenoShim1.Deno.readTextFile(); const denoShim = {};"
+    ),
+  )])
+  .await;
 }
 
 #[tokio::test]
 async fn transform_global_this_deno() {
-  let result = TestBuilder::new()
-    .with_loader(|loader| {
-      loader.add_local_file("/mod.ts", r#"globalThis.Deno.readTextFile(); globalThis.test = 5;"#);
-    })
-    .shim_package_name("test-shim")
-    .transform()
-    .await
-    .unwrap();
-
-  assert_files!(
-    result.main.files,
-    &[(
-      "mod.ts",
+  assert_transforms(vec![
+    (
+      "globalThis.Deno.readTextFile(); globalThis.test = 5;",
       concat!(
         r#"import * as denoShim from "test-shim";"#,
-        "\n({ Deno: denoShim.Deno, ...globalThis }).Deno.readTextFile(); globalThis.test = 5;"
+        "\n({ ...denoShim, ...globalThis }).Deno.readTextFile(); globalThis.test = 5;"
       )
-    )]
-  );
+    ),
+  ]).await;
+}
+
+#[tokio::test]
+async fn no_shim_for_declarations() {
+  assert_identity_transforms(vec![
+    "const { Deno } = test;",
+    "const { asdf, ...Deno } = test;",
+    "const { Deno: test } = test;",
+    "const { test: Deno } = test;",
+    "const [Deno] = test;",
+    "const [test, ...Deno] = test;",
+    "const obj = { Deno: test };",
+    "interface Deno {}",
+    "interface Test { Deno: string; }",
+    "interface Test { Deno(): string; }",
+    "class Deno {}",
+    "class Test { Deno: string; }",
+    "class Test { Deno() {} }",
+    "const t = class Deno {};",
+    "function Deno() {}",
+    "const t = function Deno() {};",
+    "import { Deno } from 'test';",
+    "import * as Deno from 'test';",
+    "import { test as Deno } from 'test';",
+    "import { Deno as test } from 'test';",
+    "export { Deno } from 'test';",
+    "export * as Deno from 'test';",
+    "export { test as Deno } from 'test';",
+    "export { Deno as test } from 'test';",
+    "try {} catch (Deno) {}",
+    "function test(Deno) {}",
+  ])
+  .await;
 }
 
 #[tokio::test]
 async fn transform_deno_collision() {
-  let result = TestBuilder::new()
-    .with_loader(|loader| {
-      loader.add_local_file(
-        "/mod.ts",
-        concat!(
-          "const Deno = {};",
-          "const { Deno: Deno2 } = globalThis;",
-          "Deno2.readTextFile();",
-          "Deno.test;"
-        ),
-      );
-    })
-    .shim_package_name("test-shim")
-    .transform()
-    .await
-    .unwrap();
-
-  assert_files!(
-    result.main.files,
-    &[(
-      "mod.ts",
-      concat!(
-        r#"import * as denoShim from "test-shim";"#,
-        "\nconst Deno = {};",
-        "const { Deno: Deno2 } = ({ Deno: denoShim.Deno, ...globalThis });",
-        "Deno2.readTextFile();",
-        "Deno.test;"
-      )
-    )]
-  );
+  assert_transforms(vec![(
+    concat!(
+      "const Deno = {};",
+      "const { Deno: Deno2 } = globalThis;",
+      "Deno2.readTextFile();",
+      "Deno.test;"
+    ),
+    concat!(
+      r#"import * as denoShim from "test-shim";"#,
+      "\nconst Deno = {};",
+      "const { Deno: Deno2 } = ({ ...denoShim, ...globalThis });",
+      "Deno2.readTextFile();",
+      "Deno.test;"
+    ),
+  )])
+  .await;
 }
 
 #[tokio::test]
@@ -171,7 +195,10 @@ async fn transform_remote_files() {
       loader
         .add_local_file(
           "/mod.ts",
-          "import * as other from 'http://localhost/mod.ts';",
+          concat!(
+            "import * as other from 'http://localhost/mod.ts';\n",
+            "import 'https://deno.land/std@0.102.0/mod.ts';",
+          ),
         )
         .add_remote_file(
           "http://localhost/mod.ts",
@@ -204,6 +231,10 @@ async fn transform_remote_files() {
           "import * as localhost2 from 'http://localhost2';",
           &[("content-type", "application/javascript")],
         )
+        .add_remote_file(
+          "https://deno.land/std@0.102.0/mod.ts",
+          "console.log(5);",
+        )
         .add_remote_file_with_headers(
           "http://localhost2",
           "import * as localhost3Mod from 'https://localhost3/mod.ts';",
@@ -228,7 +259,10 @@ async fn transform_remote_files() {
     &[
       (
         "mod.ts",
-        "import * as other from './deps/localhost/mod.js';"
+        concat!(
+          "import * as other from './deps/localhost/mod.js';\n",
+          "import './deps/deno_land/std_0.102.0/mod.js';",
+        )
       ),
       (
         "deps/localhost/mod.ts",
@@ -258,6 +292,7 @@ async fn transform_remote_files() {
         "deps/localhost/sub/subfolder.js",
         "import * as localhost2 from '../../localhost2.js';"
       ),
+      ("deps/deno_land/std_0.102.0/mod.ts", "console.log(5);"),
       (
         "deps/localhost2.js",
         "import * as localhost3Mod from './localhost3/mod.js';"
@@ -665,8 +700,12 @@ async fn transform_specifier_mappings() {
           "import * as myOther from './other.ts';",
         );
     })
-    .add_specifier_mapping("http://localhost/mod.ts", "remote-module")
-    .add_specifier_mapping("file:///file.ts", "local-module")
+    .add_specifier_mapping(
+      "http://localhost/mod.ts",
+      "remote-module",
+      Some("1.0.0"),
+    )
+    .add_specifier_mapping("file:///file.ts", "local-module", None)
     .transform()
     .await
     .unwrap();
@@ -677,6 +716,13 @@ async fn transform_specifier_mappings() {
       ("mod.ts", "import * as remote from 'remote-module';\nimport * as local from 'local-module';\n"),
     ]
   );
+  assert_eq!(
+    result.main.dependencies,
+    &[Dependency {
+      name: "remote-module".to_string(),
+      version: "1.0.0".to_string(),
+    },]
+  );
 }
 
 #[tokio::test]
@@ -685,8 +731,8 @@ async fn transform_not_found_mappings() {
     .with_loader(|loader| {
       loader.add_local_file("/mod.ts", "test");
     })
-    .add_specifier_mapping("http://localhost/mod.ts", "local-module")
-    .add_specifier_mapping("http://localhost/mod2.ts", "local-module2")
+    .add_specifier_mapping("http://localhost/mod.ts", "local-module", None)
+    .add_specifier_mapping("http://localhost/mod2.ts", "local-module2", None)
     .transform()
     .await
     .err()
@@ -866,20 +912,100 @@ async fn test_entry_points() {
 
   assert_files!(
     result.main.files,
-    &[
-      (
-        "mod.ts",
-        "import package1 from 'preact';\n",
-      )
-    ]
+    &[("mod.ts", "import package1 from 'preact';\n",)]
   );
   assert_eq!(
     result.main.dependencies,
+    &[Dependency {
+      name: "preact".to_string(),
+      version: "^10.5.0".to_string(),
+    },]
+  );
+  assert_eq!(result.main.entry_points, &["mod.ts"]);
+  assert_eq!(result.main.shim_used, false);
+
+  assert_files!(
+    result.test.files,
+    &[(
+      "mod.test.ts",
+      concat!(
+        "import * as denoShim from \"deno.ns\";\n",
+        "import './mod.js';\n",
+        "import package1 from 'preact';\n",
+        "import package3 from 'react';\n",
+        "denoShim.Deno.writeTextFile('test', 'test')"
+      ),
+    )]
+  );
+  assert_eq!(
+    result.test.dependencies,
+    &[Dependency {
+      name: "react".to_string(),
+      version: "17.0.2".to_string(),
+    },]
+  );
+  assert_eq!(result.test.entry_points, &["mod.test.ts"]);
+  assert_eq!(result.test.shim_used, true);
+}
+
+#[tokio::test]
+async fn test_entry_points_same_module_multiple_places() {
+  let result = TestBuilder::new()
+    .with_loader(|loader| {
+      loader
+        .add_local_file(
+          "/mod.ts",
+          concat!(
+            "export * from 'https://deno.land/std@0.102.0/path.ts';\n",
+            "import * as deps from './deps.ts';",
+          ),
+        )
+        // ensure that the path.ts in this file being already analyzed
+        // doesn't cause flags.ts to not be analyzed
+        .add_local_file(
+          "/deps.ts",
+          concat!(
+            "export * from 'https://deno.land/std@0.102.0/path.ts';\n",
+            "export * from 'https://deno.land/std@0.102.0/flags.ts';",
+          ),
+        )
+        .add_remote_file(
+          "https://deno.land/std@0.102.0/flags.ts",
+          "export class Flags {}",
+        )
+        .add_remote_file(
+          "https://deno.land/std@0.102.0/path.ts",
+          "export class Path {}",
+        )
+        .add_local_file("/mod.test.ts", "import * as deps from './deps.ts';");
+    })
+    .add_test_entry_point("file:///mod.test.ts")
+    .transform()
+    .await
+    .unwrap();
+
+  assert_files!(
+    result.main.files,
     &[
-      Dependency {
-        name: "preact".to_string(),
-        version: "^10.5.0".to_string(),
-      },
+      (
+        "mod.ts",
+        concat!(
+          "export * from './deps/deno_land/std_0.102.0/path.js';\n",
+          "import * as deps from './deps.js';",
+        )
+      ),
+      (
+        "deps.ts",
+        concat!(
+          "export * from './deps/deno_land/std_0.102.0/path.js';\n",
+          "export * from './deps/deno_land/std_0.102.0/flags.js';",
+        )
+      ),
+      (
+        "deps/deno_land/std_0.102.0/flags.ts",
+        "export class Flags {}"
+      ),
+      ("deps/deno_land/std_0.102.0/path.ts", "export class Path {}")
     ]
   );
   assert_eq!(result.main.entry_points, &["mod.ts"]);
@@ -887,28 +1013,8 @@ async fn test_entry_points() {
 
   assert_files!(
     result.test.files,
-    &[
-      (
-        "mod.test.ts",
-        concat!(
-          "import * as denoShim from \"deno.ns\";\n",
-          "import './mod.js';\n",
-          "import package1 from 'preact';\n",
-          "import package3 from 'react';\n",
-          "denoShim.Deno.writeTextFile('test', 'test')"
-        ),
-      )
-    ]
-  );
-  assert_eq!(
-    result.test.dependencies,
-    &[
-      Dependency {
-        name: "react".to_string(),
-        version: "17.0.2".to_string(),
-      },
-    ]
+    &[("mod.test.ts", "import * as deps from './deps.js';",)]
   );
   assert_eq!(result.test.entry_points, &["mod.test.ts"]);
-  assert_eq!(result.test.shim_used, true);
+  assert_eq!(result.test.shim_used, false);
 }
